@@ -1,3 +1,4 @@
-if __name__ == '__main__':
+if __name__ == "__main__":
     from .cli import main
+
     main(as_module=True)
